#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>

void to_kill();
pid_t pid;

int
main (int argc, char **argv)
{
    int c;
    extern char *optarg;
    extern int optind;
    char *timeout = NULL;
    int err = 0;

    struct sigaction sact;
    struct sigaction * osact = NULL;
    int status, exit_status, signal_status;

    while ((c = getopt(argc, argv, "t:")) != EOF)
        switch (c) {
        case 't':
            timeout = optarg;
            break;
        case '?': 
            err++;
        }

    if ( err > 0 || argc < 4 ) {
        printf("usage: timeout [-t <second>] unix command with arguments\n");
        printf("\n");
        printf("after <second> seconds, timeout kills unix command if it\n");
        printf("has not finished. if <second>=0, timeout does not kill.\n");
        printf("\n");
        printf("Send comments to:\n");
        printf("Michael Wang, mwang@mindspring.com, Thanks.\n");
        printf("\n");
        printf("version: 1.0, Feburary, 1998.\n");
        exit (-100);
    }

    sact._funcptr._handler = to_kill;
    sact.sa_flags = SA_RESTART;
    sigemptyset(&sact.sa_mask);

    if ( pid = fork() ) {
        if ( sigaction(SIGALRM, &sact, osact) < 0 ) {
            fprintf (stderr, "can not set signal handler for SIGTALRM.\n");
            exit(-100);
        }
        alarm( atoi(timeout) );
        waitpid (pid, &status, 0);
        alarm(0);        /* disable alarm */
        if ( WIFEXITED(status) ) {
            exit_status = WEXITSTATUS(status); 
            printf ("normal exit, status=%d.\n", exit_status);
            exit (exit_status);
        } else if ( WIFSIGNALED(status) ) {
            signal_status = WTERMSIG(status); 
            printf ("killed, signal=%d (%s)", signal_status, 
                     strsignal(signal_status));
            if ( WCOREDUMP(status) )
                printf(", and coredumped.\n");
            else
                printf(".\n");
            exit (-signal_status);
        } 
    } else if ( pid == (pid_t) 0 ) {
        execvp ( argv[optind], argv+optind );
    }
}

void
to_kill()
{
   if ( kill(pid, SIGTERM) == 0 ) {        /* use normal kill first */
       printf ("kill with SIGTERM, ");
   } else if ( kill(pid, SIGKILL) == 0 ) { /* use kill -9 */
       printf ("kill with SIGKILL, ");
   }
}
